/**
 * Alibaba webRTC JS SDK
 * @autor gaosong
 * @verson 1.0.0
 * @since 2016-03-21
 */
import ConfigManager from "./ConfigManager";
import Signaling from "./Signaling";
import Session from "./Session";
import StateMachine from "./StateMachine";
import Monitor from "./Monitor";
import {EndReason,CallType} from "./Enum";

ConfigManager.setDefaultConfig('maxRecallCount', 15);
let _delayCallTimer;
/**
 *
 * artc是应用层抽象,包含视频通话的所有业务行为
 *
 * artc状态机(状态机原理参见StateMachine.js):
 *               offline[离线状态] 这 该状态会尝试连接信令服务器 连接成功后转入idle状态
 *                      ^
 *                      |
 *                      v
 * new[初始状态]-->idle [空闲(在线)状态] 此状态会监听来电 这个过程中如果websocket异常会转入offline状态
 *                      ^
 *                      |
 *                      v
 *               session会话状态 此状态对应dialog的整个状态机流转
 * */
class ARTC extends StateMachine {
    constructor(config) {
        //初始化状态机构造器
        super('ARTC');
        ConfigManager.init(config);
        let artc = this;
        artc.retryCount = 0;
        window.__artc = this;//for debug
        //如果artc 在new或者offline状态超过了60秒则触发timeout事件
        artc.timeout('new,offline', 60, function () {
            artc.emit('socketTimeout');
        });
        this.signaling = new Signaling();
        //创建信令服务
        this.connectSignaling(false);
        this.on('dialog.connected', function () {
            artc.retryCount = 0;
        })
    }

    //空闲状态
    onIdle(reonline) {
        let artc = this;
        //空闲状态时 释放掉上一次会话
        if (artc.dialog && !artc.dialog.inConnected()) {
            artc.dialog = null;
        }
        //空闲等待呼叫
        this.signaling.waitCall(function (response) {
            if (response.type == CallType.CloseSession) {//closeCall
                if (artc.dialog && artc.dialog.isValidResponse(response)) {
                    artc.dialog.hangup(EndReason.ThatHangup);
                }
            }
            else {
                if (artc.dialog) {
                    //来电话了 如果已经有会话 而且新的call跟当前的会话一致
                    if (artc.dialog.isValidResponse(response)) {
                        //挂断当前的会话
                        artc.dialog.hangup(EndReason.ThatRecall);
                    }
                    else {
                        //回复忙
                        artc.signaling.sayBusy(response.initiativeUserInfo, response.passtiveUserInfo);
                        return;
                    }
                }
                //创建会话
                artc.dialog = new Session(artc.signaling, artc);
                //同步事件 @see stateMachine里的注释
                //同步状态 @see stateMachine里的注释
                //当dialog转入end状态时 artc则转入idle状态
                artc.syncState(artc.dialog, 'end', 'idle');

                artc.dialog.incomingCall(response);

                if (response.type == CallType.AutomaticSession) {
                    artc.dialog.answer();
                }
                artc.nextState('session');
            }
        });
        if (this._call) {
            this._call();
            this._call = null;
        }
        else if (reonline === true) {
            this.autoCall();
        }

    }

    onOffline() {
        //this.connectSignaling(true);

    }

    connectSignaling(reonline) {
        let artc = this;
        artc.signaling.connect().then(function () {
            artc.signaling.retryCount = 0;
            artc.nextState('idle', reonline);
            artc.emit('online');
        }).catch(function (socketError) {
            if (artc.signaling.retryCount > 5) {
                artc.emit('socketError');
            }
            //信令连接(websocket)异常时 artc转入offline状态
            if (artc.dialog && !artc.dialog.inConnected()) {
                artc.dialog.close();
            }
            clearTimeout(artc._retryTimer);
            artc._retryTimer = setTimeout(function () {
                artc.signaling.retryCount++;
                artc.connectSignaling(true);
                artc.nextState('offline');
            }, socketError ? 3000 : 0);
        });
    }

    autoCall(allowVideo) {
        let sessionInfo;
        if (sessionInfo = sessionStorage.getItem('artc_dialog')) {

            sessionInfo = JSON.parse(sessionInfo);
            if (allowVideo !== undefined) {
                sessionInfo.allowVideo = allowVideo;
            }
            if (this.retryCount++ < ConfigManager.get('maxRecallCount')) {
                Monitor.log('第' + this.retryCount + '次autoCall');
                this.call(sessionInfo.initiativeUserInfo, sessionInfo.passtiveUserInfo, CallType.AutomaticSession, sessionInfo.allowVideo);
            }
            else {
                this.emit('recallFailed');
                console.log('已达最大重试次数')
            }
        }
    }

    call(initiativeUserInfo, passtiveUserInfo, type = CallType.CoupleSession, allowVideo = true) {
        let artc = this;
        if (arguments.length == 1) {
            passtiveUserInfo = initiativeUserInfo.passtiveUserInfo;
            type = initiativeUserInfo.type;
            allowVideo = initiativeUserInfo.allowVideo;
            initiativeUserInfo = initiativeUserInfo.initiativeUserInfo;
        }
        function _call() {
            if (artc.dialog) {
                if (artc.dialog.isSameSession(initiativeUserInfo, passtiveUserInfo)) {
                    artc.dialog.hangup(EndReason.ThisRecall);
                    artc.dialog = null;
                    clearTimeout(_delayCallTimer);
                    _delayCallTimer = setTimeout(function () {
                        _call();
                    }, 2500);
                    return;
                }
                else {
                    console.log('already in call');
                }
            }
            if (!artc.dialog) {
                artc.dialog = new Session(artc.signaling, artc);
                artc.syncState(artc.dialog, 'end', 'idle');
                artc.nextState('session');
                artc.dialog.call(initiativeUserInfo, passtiveUserInfo, type, allowVideo);

                sessionStorage.setItem('artc_dialog', JSON.stringify({
                    initiativeUserInfo: initiativeUserInfo,
                    passtiveUserInfo: passtiveUserInfo,
                    allowVideo: allowVideo
                }))
            }
        }

        if (this.signaling.isSocketReady()) {
            _call();
        }
        else {
            this._call = _call;
        }
    }

    hangup(endReason = EndReason.ThisHangup) {
        if (this.dialog) {
            this.dialog.hangup(endReason);

        }
    }

    recall() {
        let session = sessionStorage.getItem('artc_dialog');
        if (session) {
            session = JSON.parse(session);
            this.call(session.initiativeUserInfo, session.passtiveUserInfo);
        }
    }

    answer() {
        if (this.dialog && this.dialog.isInState('incomingCall,ready')) {
            this.dialog.answer();
        }
    }

    toggleVideoMute() {
        if (this.dialog && this.dialog.localMedia) {
            return this.dialog.localMedia.toggleVideoMute();
        }
    }

    isVideoMute() {
        if (this.dialog && this.dialog.localMedia) {
            return this.dialog.localMedia.isVideoMute();
        }
        return false;
    }

    setReporter(reporter) {
        Monitor.setReporter(reporter);
    }

    setTagReporter(reporter) {
        Monitor.setTagReporter(reporter);
    }

    switchSignaling(url) {
        ConfigManager.set('url', url);
        clearTimeout(this._retryTimer);
        if (this.signaling.socket.readyState == WebSocket.OPEN) {
            this.signaling.socket.close();
        }
        else {
            if (this.dialog && !this.dialog.inConnected()) {
                this.dialog.close();
            }
            this.connectSignaling(true);
            this.nextState('offline');
        }
    }

    on(method, callback) {
        let mapper;
        if (mapper = _eventMapper[method]) {
            mapper.split(',').forEach(function (m) {
                super.on(m, callback);
            });

        }
        else {
            super.on(method, callback);
        }
    }
}
const _eventMapper = {
    'calling': 'dialog.calling',
    'connect': 'online',
    'notConnect': 'offline',
    'connected': 'dialog.connected',
    'busy': 'dialog.busy',
    'comingCall': 'dialog.incomingCall',
    'notInroom': 'dialog.offline',
    'hangup': 'dialog.hangup',
    'onFailed': 'dialog.failed',
    'stats': 'dialog.peer.stats',
    'error': 'dialog.error',
    'socketError': 'dialog.socketError',
    'averageStats': 'dialog.peer.averageStats',
    'mediaDenied': 'dialog.mediaError',
    'opposite_offline': 'dialog.disconnected',
    'timeout': 'timeout,dialog.timeout',
    'active': 'dialog.connected',
    'wakeCam': 'dialog.mediaPermit',
    'iOS-online': 'dialog.iOS-online'
};

window.ARTC = ARTC;
ARTC.ConfigManager = ConfigManager;
ARTC.Monitor = Monitor;
ARTC.EndReason = EndReason;
ARTC.CallType = CallType;



