/**
 * Created by godsong on 16/4/21.
 */

const LOCAL_STORAGE_NAME = 'artc_logs';
let rFormat = /%[sdifoOc]/g;
let _uuid = 0;
class Logger {
    constructor(name, size, delay) {
        this.localStorageKey = 'logs_' + name;
        let logBuffer = localStorage.getItem(this.localStorageKey);
        if (logBuffer) {
            try {
                logBuffer = JSON.parse(logBuffer);
            } catch (e) {
                console.warn();
            }

        }
        this.logBuffer = logBuffer instanceof Array ? logBuffer : [];
        if (size < 300) {
            console.warn('log size limit must above 300. now it set to 300');
            size = 300;
        }
        this.sizeLimit = size || 1024;
        this.logBufferSize = 0;
        this.reporterManager = new ReporterManager(name, delay || 0);
        let self = this;
        if (this.logBuffer.length > 0) {
            clearTimeout(this._reportTimeout);
            this._reportTimeout = setTimeout(function () {
                self._doReport();
                self.updateStorage();
            }, 5000);
        }
    }

    updateStorage() {
        localStorage.setItem(this.localStorageKey, JSON.stringify(this.logBuffer));
    }

    log() {
        this._log(Array.prototype.slice.call(arguments));
    }

    trace() {
        this._log(Array.prototype.slice.call(arguments), 'trace');
    }

    debug() {
        this._log(Array.prototype.slice.call(arguments), 'debug');
    }

    warn() {
        this._log(Array.prototype.slice.call(arguments), 'warn');
    }

    error() {
        this._log(Array.prototype.slice.call(arguments), 'error');
    }

    setReporter(reporter) {
        this.reporterManager.setReporter(reporter);

    }

    _log(args, level = 'log') {
        let logText = this._buildLogText(args);
        if (!this._isRepeat(logText)) {
            args[0] = '>' + _time() + '\n' + args[0];
            console[level].apply(console, args);
            this.reportLog(logText);
        }
        else {
            console.log(logText);
        }
        return logText;
    }

    _split(text, limit, refactor) {
        let pointer = limit / refactor | 0;
        let safe = 0;
        let slice;
        while (safe++ < 100) {
            slice = text.substr(0, pointer);
            let len = encodeURIComponent(slice).length;
            let dist = Math.floor((limit - len) / 9);
            if (dist == 0) {
                break;
            }
            else {
                pointer += dist;
            }
        }
        return [slice, text.slice(pointer)];
    }

    reportLog(logText, id) {
        let encode = encodeURIComponent(logText);
        let otherLen = 140;//140是logObj其余部分的长度
        let len = encode.length + otherLen;
        if (len > this.sizeLimit) {
            let split;
            if (this.sizeLimit - this.logBufferSize - otherLen > this.sizeLimit * 0.2) {
                split = this._split(logText, this.sizeLimit - this.logBufferSize - otherLen)
            }
            else {
                split = this._split(logText, this.sizeLimit - otherLen);
            }
            this.reportLog(split[0], id || ++_uuid);
            this.reportLog(split[1], id || _uuid);
            return;
        }
        let logObj = {
            time: _time(),
            count: 1,
            log: logText
        };
        if (id > 0) {
            logObj.id = id;
        }
        if (len + this.logBufferSize > this.sizeLimit) {
            this._doReport()
        }
        this.logBuffer.push(logObj);
        this.logBufferSize += len;
        let self = this;
        clearTimeout(this._reportTimeout);
        this._reportTimeout = setTimeout(function () {
            self._doReport();
            self.updateStorage();
        }, 5000);
        this.updateStorage();
    }

    reportData(data) {
        this.reporterManager.add(data);
    }

    _doReport() {
        this.reporterManager.add(this.logBuffer);
        this.logBuffer = [];
        this.logBufferSize = 0;
    }

    _buildLogText(args) {
        let logger = this,
            logText = logger._serialize(args[0]),
            pt = 1;
        logText = logText.replace(rFormat, function (m) {
            if (args.length > 0) {
                return logger._serialize(args[pt++]);
            }
            else return m;
        });
        for (; pt < args.length; pt++) {
            logText += ' ' + logger._serialize(args[pt]);
        }
        return logText;
    }

    _isRepeat(logText) {
        let prev = this.logBuffer[this.logBuffer.length - 1];
        if (prev && prev.log == logText) {
            prev.count++;
            prev.endTime = _time();
            this.updateStorage();
            return true;
        }
        else {
            if (prev && prev.count == 1) {
                delete prev.count;
            }
            return false;
        }
    }


    _serialize(obj) {
        let serialized = this.customSerializer && this.customSerializer(obj);
        if (serialized === undefined) {
            if (obj instanceof Error) {
                serialized = '[' + obj.name + ':' + obj.message + ']';
            }
            if (typeof obj === 'object') {
                serialized = JSON.stringify(obj);
            }
            else {
                serialized = obj.toString();
            }
        }
        return serialized;

    }

}
function _time() {
    let now = new Date();
    return now.getHours() + ':' + now.getMinutes() + ':' + now.getSeconds() + '.' + now.getMilliseconds();
}
class ReporterManager {
    constructor(name, delay) {
        this.localStorageKey = 'reporter_queue_' + name;
        let reporterQueue = localStorage.getItem(this.localStorageKey);
        if (reporterQueue) {
            try {
                reporterQueue = JSON.parse(reporterQueue)
            } catch (e) {
                console.warn('reporter queue local storage error!');
            }
        }
        this.onLineCount = 0;
        this.onlineDetector = -1;
        this.reporterQueue = reporterQueue instanceof Array ? reporterQueue : [];
        this.delay = delay;
        this.lastReportTime = 0;
        this.reporting = false;
    }

    setReporter(reporter) {
        this.reporter = reporter;
        this.needToReport();
    }

    updateStorage() {
        localStorage.setItem(this.localStorageKey, JSON.stringify(this.reporterQueue));
    }

    needToReport() {
        let self = this;
        if (this.reporterQueue.length > 0) {
            if (Logger.onLine) {
                let now = +new Date();
                if (now - this.lastReportTime >= this.delay || now - this.lastReportTime < 80) {
                    this.lastReportTime = now;
                    this.reporter(this.reporterQueue.shift());
                    this.updateStorage();
                    if (this.reporterQueue.length > 0) {
                        this.needToReport();
                    }
                    else {
                        this.reporting = false;
                    }
                }
                else {
                    this.reporting = true;
                    setTimeout(function () {
                        self.needToReport();
                    }, this.delay - (now - this.lastReportTime));
                }
            }
            else {
                if (this.onlineDetector == -1) {
                    this.reporting = true;
                    this.onlineDetector = setInterval(function () {
                        if (Logger.onLine) {
                            self.onLineCount++;
                        }
                        if (self.onLineCount >= 3) {
                            clearInterval(self.onlineDetector);
                            self.onlineDetector = -1;
                            self.onLineCount = 0;
                            self.needToReport();
                        }
                    }, 1000);
                }
            }
        }
    }

    add(reportItem) {
        this.reporterQueue.push(reportItem);
        this.updateStorage();
        if (!this.reporting && this.reporter) {
            this.needToReport();
        }
    }

}
Object.defineProperty(Logger, 'onLine', {
    get: function () {
        if (this._mock) {
            return this._mock.value;
        }
        else {
            return navigator.onLine;
        }
    },
    set: function (mock) {
        if (typeof mock === 'boolean') {
            this._mock = {value: mock};
        }
        else if (mock === null) {
            delete this._mock;
        }
        else {
            this._mock = mock;
        }

    }
})
export default Logger;