/**
 * Created by gaosong on 16/3/24.
 */
import StateMachine from './StateMachine';
import Signaling from './Signaling';
import ConnectionManager from './ConnectionManager';
import Media from './MediaController';
import ConfigManager from './ConfigManager';
import Monitor from "./Monitor";
import {EndReason,CallStatus} from "./Enum";

ConfigManager.setDefaultConfig('timeout', 60);
ConfigManager.setDefaultConfig('getUserMediaTimeout', 180);
let _delay1,_delay2;

/**
 * Session 抽象一次会话
 * 状态机:
 * 主叫方:call [信令呼叫状态]                   被叫方:  incomingCall [被叫状态]
 *        |   此状态会发出ｃａｌｌ信令等信令服务器返回结果     |  此状态会设置对方传过来的candidate和sdp
 *        | ----> ｅｎｄ 对方不在线进入end状态             |  等对方成功传过来sdp进入下个状态
 *        v   正常则进入下个状态                          v
 *    getUserMedia[获取用户本地视频状态]                ready [可以接听状态]
 *        |   拿到用户本地视频流之后进入下个状态             |   只有进入ready状态才允许接听
 *        v                                           |  因为必须等对端的sdp发过来
 *     callinｇ [ｒｔｃ呼叫状态]                         ｖ  完成setRemoteDescription
 *        |    此状态会创建ｓdp和candidate并发出        answer [接听状态]
 *        |   等createOffer完成并把ｓｄｐ发出后 进入下个状态 |  点击接听后进入接听状态
 *        v                                           |  此状态会收集candidate
 *     waitAnswer [等待接听状态] 此状态监听对方的响应       ｖ
 *        |   拿到对方响应(ＲｅｓＣａｌｌ)后进入下个状态    getUserMedia[获取用户本地视频状态]
 *        |                                           |
 *        v                                           v
 *     receiveAnswer [处理响应状态]                   accept [接受状态]
 *        | ----> end 对方拒绝或者忙                     ^
 *        |    对方应答则开始进行ｉｃe打洞建联              |   此状态createAnswer并发送accept信令
 *        v                                           v   然后开始ice打洞建联
 *                  connected 之后会话状态会跟PeerConnection的状态同步
 *                      ^
 *                      |
 *                      V
 *                  disconnected <->failed
 * */
export default class Session extends StateMachine {
    constructor(signaling, parent) {
        super('dialog', parent);
        let dialog = this;
        dialog.signaling = signaling;
        this.stats={
            rtt:'0/1',
            delay:'0/1',
            frameRateSent:'0/1',
            frameRateReceived:'0/1'
        };
        dialog.connectionManager = new ConnectionManager(dialog);
        dialog.localMedia = new Media(ConfigManager.get('localVideo'));
        dialog.remoteMedia = new Media(ConfigManager.get('remoteVideo'));
        dialog.signaling.off('error');
        dialog.signaling.on('error', function (error, type) {
            Monitor.log(type + ':', error);
            dialog.nextState('end', EndReason.transFromName(type));
            dialog.emitError(type, error);
        });
        dialog.connectionManager.on('error', function (error, type) {
            Monitor.log(type + ':', error);
            dialog.nextState('end', EndReason.transFromName(type));
            dialog.emitError(type, error);
        });
        dialog.signaling.returnSDPTime();
        dialog.signaling.onIPInfo(function(data){
            dialog.initiativeUserInfo.ipInfo=data.initiativeUserIpInfo;
            dialog.initiativeUserInfo.ip=data.initiativeUserIp;
            dialog.passtiveUserInfo.ipInfo=data.passiveUserIpInfo;
            dialog.passtiveUserInfo.ip=data.passiveUserIp;
        });
        //todo broadcast leave的type pc端和移动端有严重出入!!!
        dialog.signaling.onLeave(function (type) {
            var endReason;
            if(type===0){
                endReason=EndReason.ThatHangup;
            }
            else if(type<0){
                endReason=EndReason.transFromCode(-type);
            }
            else if(type>1){
                endReason=EndReason.mix(EndReason.ThatError|EndReason.transFromCode(-type));
            }
            dialog.nextState('end', endReason);
        });
        dialog.signaling.onIceRestart(function (sdpString) {
            dialog.restartIce(sdpString);
        });
        dialog.timeout('getUserMedia', 1.5, function () {
            dialog.emit('mediaPermit', false);
        });
        dialog.timeout('call+getUserMedia+calling+waitAnswer+receiveAnswer', ConfigManager.get('timeout'), function () {
            dialog.emit('timeout');
        });
        dialog.timeout('peer.checking',30,function(){
            dialog.emit('timeout');
        });
        dialog.connectionManager.on('stats',function(report){
            'rtt,delay,frameRateSent,frameRateReceived'.split(',').forEach(function(prop){
                dialog.stats[prop]=dialog.stats[prop].replace(/(\d+)\/(\d+)/g,function(m,a,b){
                    return (+a+(+report[prop]))+'/'+(+b+1);
                })
            });
            dialog.stats.localCandidateType=report['localCandidateType'];
            dialog.stats.remoteCandidateType=report['remoteCandidateType'];
            dialog.stats.transportType=report['transportType'];

        });

       // dialog.signaling.emitError('socketError',new Error('jb'));
        //dialog.breakOn('*->connected')
    }

    onCall(type) {
        let dialog = this;
        dialog.signaling.call(this.initiativeUserInfo, this.passtiveUserInfo, type, this.allowVideo).then(function (response) {
            switch (response.status) {
                case CallStatus.OK:
                    dialog.roomId = response.roomId;
                    dialog.connectionManager.connect(response.turnServers,dialog.allowVideo);
                    dialog.remoteMedia.fetchMedia(dialog.connectionManager);
                    dialog.nextState('getUserMedia', 'calling');
                    break;
                case CallStatus.OFFLINE:
                    console.log('对方不在线');
                    dialog.emit('offline');
                    dialog.nextState('end', EndReason.ThatOffline);
                    dialog.signaling.waitIOSOnline(function(){
                        dialog.emit('iOS-online');
                    });
                    break;
                case CallStatus.ONLINE_BUSY:
                    console.log('对方占线2');
                    dialog.emit('busy');
                    dialog.nextState('end', EndReason.ThatBusy);
                    break;
                case CallStatus.REJECT:
                    console.log('对方拒绝');
                    dialog.emit('reject');
                    dialog.nextState('end', EndReason.ThatReject);
                    break;
                case CallStatus.IOS_ONLINE:
                    console.log('ios-online');
                    dialog.nextState('end',EndReason.ThatError);

                    break;
            }
        }).catch(function(e){
           throw e;
        });
    }

    onGetUserMedia(nextState) {
        let dialog = this;
        dialog.localMedia.setAudioMute(true);
        dialog.localMedia.initMedia(dialog.allowVideo);
        dialog.localMedia.fetchMedia();
        dialog.localMedia.sendTo(dialog.connectionManager.peerConnection).then(function (mediaStream) {
            /*if (dialog.elapse() > ConfigManager.get('getUserMediaTimeout')) {
                dialog.emit('getUserMediaTimeout');
            }*/
            Monitor.log('mediaReady');
            window._ms = mediaStream;
            dialog.emit('userMediaReady', dialog.elapse() > ConfigManager.get('getUserMediaTimeout'));
            dialog.nextState(nextState);
        }).catch(function (e) {
            console.log(e);
            dialog.emitError('mediaError', e);
            dialog.nextState('end', EndReason.transFromName('mediaError'));
        });

    }

    onCalling() {

        let dialog = this;

        dialog.connectionManager.createCandidate(function (candidate) {
            dialog.signaling.sendCandidate(candidate, dialog.roomId, dialog.initiativeUserInfo);
        });
        dialog.connectionManager.addCandidate(dialog.signaling.getCandidate(), dialog.thatUserInfo.userId);
        dialog.connectionManager.createOffer(function (sdp) {
            dialog.signaling.sendSDP(sdp.sdp, dialog.roomId, dialog.initiativeUserInfo);
            dialog.nextState('waitAnswer');
        });
        Monitor.addTag('DIALOG_NEW',{
            roomId:dialog.roomId,
            role:dialog.role,
            thisUserInfo:dialog.thisUserInfo,
            thatUserInfo:dialog.thatUserInfo
        });

    }

    onWaitAnswer() {
        let dialog = this;
        dialog.signaling.waitAnswer(function (response) {
            dialog.nextState('receiveAnswer', response);
            _setSessionUserInfo(dialog, response.initiativeUserInfo, response.passtiveUserInfo, true);
            clearTimeout(_delay1);
            _delay1=setTimeout(function(){
                dialog.signaling.sendIpInfo(response.initiativeUserInfo,response.passtiveUserInfo)
            },300)
        })
    }

    onReceiveAnswer(response) {
        let dialog = this;
        switch (response.status) {
            case CallStatus.OK:
                dialog.connectionManager.setRemoteSDP(response.sdpInfo);
                dialog.syncState(dialog.connectionManager, 'connected');
                dialog.syncState(dialog.connectionManager, 'disconnected');
                dialog.syncState(dialog.connectionManager, 'failed');
                dialog.syncState(dialog.connectionManager, 'completed', 'connected');
                break;
            case CallStatus.ONLINE_BUSY:
                console.log('对方占线');
                dialog.emit('busy');
                dialog.nextState('end', EndReason.ThatBusy);
                break;
            case CallStatus.REJECT:
                console.log('对方拒绝');
                dialog.emit('reject');
                dialog.nextState('end', EndReason.ThatReject);
                break;
            case CallStatus.IOS_ONLINE:
                console.log('ios-online');
                dialog.nextState('end',EndReason.ThatError);
                dialog.emit('iOS-online');
                break;
        }
    }

    onIncomingCall() {
        let dialog = this;
        dialog.connectionManager.addCandidate(dialog.signaling.getCandidate(), dialog.thatUserInfo.userId);
        dialog.connectionManager.setSDP(dialog.signaling.getSDP());
        dialog.remoteMedia.fetchMedia(dialog.connectionManager);
        dialog.nextStateWhen(dialog.connectionManager.remoteSDPReady(), 'ready');
    }

    onReady() {
        if (this._answered) {
            this.nextState('answer');
        }
    }

    onAnswer() {
        let dialog = this;
        dialog.connectionManager.createCandidate(function (candidate) {
            dialog.signaling.sendCandidate(candidate, dialog.roomId, dialog.thisUserInfo);
        });
        dialog.nextState('getUserMedia', 'accept');
    }

    onAccept() {
        let dialog = this;
        dialog.connectionManager.createAnswer(function (sdp) {
            dialog.signaling.accept(sdp.sdp, dialog.roomId, dialog.initiativeUserInfo, dialog.passtiveUserInfo);
        });
        dialog.syncState(dialog.connectionManager, 'connected');
        dialog.syncState(dialog.connectionManager, 'disconnected');
        dialog.syncState(dialog.connectionManager, 'completed', 'connected');
        dialog.syncState(dialog.connectionManager, 'failed');
    }

    onHangup() {
        let dialog = this;
        dialog.signaling.leaveRoom(dialog.roomId, dialog.thisUserInfo, 0);
        sessionStorage.removeItem('artc_dialog');
        dialog.nextState('end', EndReason.ThisHangup);
    }


    onFailed() {
        this.nextState('end', EndReason.Failed);
    }

    onEnd(endReason) {
        this.localMedia && this.localMedia.close();
        this.localMedia && this.remoteMedia.close();
        this.signaling.reset();
        this.localMedia = null;
        this.remoteMedia = null;
        this.connectionManager&&this.connectionManager.close();
        this.connectionManager = null;
        if(endReason.is(EndReason.ThisHangup|EndReason.ThatHangup)){
            sessionStorage.removeItem('artc_dialog');
        }
        if (endReason.is(EndReason.ThisHangup|EndReason.ThisRecall)||this.isOnlyThisError(endReason)) {
            if(this.roomId) {
                if(endReason==EndReason.ThisRecall){
                    var type=-EndReason.ThatRecall.code;
                }
                else{
                    type=-endReason.code;
                }
                this.signaling.leaveRoom(this.roomId, this.thisUserInfo,type);
            }
        }
    }
    isOnlyThisError(endReason){
        return endReason.code < 0 && endReason.not(EndReason.Failed|EndReason.ThatError|EndReason.SocketError);
    }
    isErrorEndReason(endReason){
        return endReason.code<0||endReason.has('ThatError');
    }
    afterEnd(endReason) {
        this.destroy();
        this.emit('end.'+endReason.name);
        Monitor.log('dialog end with reason:' + endReason.name);
        Monitor.addTag('DIALOG_END',{
            reason:endReason.name,
            hasError:this.isErrorEndReason(endReason),
            rtt:eval(this.stats.rtt).toFixed(2),
            delay:eval(this.stats.delay).toFixed(2),
            localType:this.stats.localCandidateType,
            remoteType:this.stats.remoteCandidateType,
            fps:eval(this.stats.frameRateSent).toFixed(2),
            elapse:this.elapse('connected').toFixed(3),
            thisUserInfo:this.thisUserInfo,
            thatUserInfo:this.thatUserInfo,
            roomId:this.roomId
        });
    }
    onConnected(){
        Monitor.addTag('DIALOG_CONNECTED',{
            elapse:this.elapse('new+cal+calling+receiveAnswer+incomingCall+ready+answer+accept').toFixed(3),
            waitAnswer:this.elapse('waitAnswer'),
            checking:this.connectionManager.elapse('checking'),
            role:this.role,
            roomId:this.roomId,
            thisUserInfo:this.thisUserInfo,
            thatUserInfo:this.thatUserInfo
        });
    }
    incomingCall(response) {
        let dialog = this;
        dialog.role="callee";
        dialog.roomId = response.roomId;

        _setSessionUserInfo(dialog, response.initiativeUserInfo, response.passtiveUserInfo, false);
        Monitor.addTag('DIALOG_NEW',{
            roomId:dialog.roomId,
            role:dialog.role,
            thisUserInfo:dialog.thisUserInfo,
            thatUserInfo:dialog.thatUserInfo

        });
        dialog.allowVideo = response.allowVideo;
        dialog.connectionManager.connect(response.turnServers,response.allowVideo);
        dialog.nextState('incomingCall');

        return dialog;
    }

    call(initiativeUserInfo, passtiveUserInfo, type, allowVideo) {
        let dialog = this;
        dialog.role="caller";
        _setSessionUserInfo(dialog, initiativeUserInfo, passtiveUserInfo, true);
        this.allowVideo = allowVideo;
        dialog.nextState('call', type);
        return dialog;

    }

    recall() {
        this.nextState('call');
    }

    answer() {
        if (this.currentState != 'ready') {
            this._answered = true;
        }
        else {
            this.nextState('answer');
        }
    }

    hangup(endReason=EndReason.ThisHangup) {

        if (this.isInState('incomingCall,ready')&&endReason==EndReason.ThisHangup) {
            this.reject();
        }
        else if (this.isInState('call,calling,waitAnswer')) {
            if(endReason===EndReason.ThisRecall){
                if(this.currentState==='call'){
                    this.onCalling=function(){
                        let dialog=this;
                        clearTimeout(_delay2);
                        _delay2=setTimeout(function(){
                            dialog.nextState('end', endReason);
                        })

                    };
                }
                else{
                    this.nextState('end', endReason);
                }
            }
            else {
                this.closeCall(endReason);
            }
        }
        else if(this.currentState!='end') {
            this.nextState('end', endReason);
        }
    }

    isSameSession(initiativeUserInfo, passtiveUserInfo) {
        return initiativeUserInfo.userId == this.initiativeUserInfo.userId && passtiveUserInfo.userId == this.passtiveUserInfo.userId;
    }

    isValidResponse(response) {

        if (response.userId) {
            return response.userId == this.thatUserInfo.userId;
        }
        else {
            return response.initiativeUserInfo.userId == this.initiativeUserInfo.userId && this.passtiveUserInfo.userId == response.passtiveUserInfo.userId;
        }

    }

    restartIce(sdpString) {
        Monitor.log('restart ice');
        this.connectionManager.setRemoteSDP({
            type: "offer",
            sdp: sdpString
        });
    }
    inConnected(){
        return this.currentState=='connected'&&this.connectionManager.isInState('connected,completed');
    }
    closeCall(endReason) {
        this.signaling.closeCall(this.initiativeUserInfo, this.passtiveUserInfo);
        this.nextState('end', endReason)
    }
    close(){
        this.localMedia && this.localMedia.close();
        this.localMedia && this.remoteMedia.close();
        this.signaling.reset();
        this.localMedia = null;
        this.remoteMedia = null;
        this.connectionManager&&this.connectionManager.close();
        this.connectionManager = null;
        this._destroy();
    }
    reject() {
        this.signaling.reject(this.initiativeUserInfo, this.passtiveUserInfo, this.roomId);
        this.nextState('end', EndReason.ThisReject);
    }
}
function _bindUA(target) {
    target.carrier = 'PC';
    target.connectionType = (navigator.connection && navigator.connection.type) || 'unknown';
    target.platform = Monitor.UA.platform;
    target.model =Monitor.UA.name;
    target.version = Monitor.UA.version;
}
function _setSessionUserInfo(target, initiativeUserInfo, passtiveUserInfo, isCaller) {
    target.initiativeUserInfo = initiativeUserInfo;
    target.passtiveUserInfo = passtiveUserInfo;
    if (isCaller) {
        target.thisUserInfo = initiativeUserInfo;
        target.thatUserInfo = passtiveUserInfo;
    }
    else {
        target.thisUserInfo = passtiveUserInfo;
        target.thatUserInfo = initiativeUserInfo;
    }
    _bindUA(target.thisUserInfo);
}



