/**
 * Created by godsong on 16/3/23.
 */
import Monitor from './Monitor';

let _stateMachines = [];
let _lastTimeStamp = +new Date();
let stateMachineTimer = setInterval(function () {
    let now = +new Date();
    let _interval = now - _lastTimeStamp;
    _lastTimeStamp = now;
    _stateMachines.forEach(function (sm) {
        sm._ticker = (now - sm._tickerStamp) / 1000;
        sm._checkTimeout(sm.currentState, _interval);
        sm.emit('timing');

    })
}, 1000);
export default class StateMachine {

    constructor(name, parent) {
        this._stateMachineName = name;
        this._eventHandlers = {};
        this.currentState = 'new';
        this.stateStack = [];
        this._breakPoint = [];
        this._parentStateMachine = parent;
        this._timeoutMonitor = [];
        this._ticker = 0;
        this._tickerStamp = +new Date();
        this._destroyed = false;
        _stateMachines.push(this);
        /* this.on('timing',function(){
         if (this.currentState !== 'new') {
         console.log(this._stateMachineName + '[' + this.currentState + '] has lasted for ' + this._ticker + 's');
         }
         })*/
    }

    on(method, handler) {
        if (!this._eventHandlers[method]) {
            this._eventHandlers[method] = [];
        }
        this._eventHandlers[method].push(handler);
    }

    log(currentState, nextState, args) {
        let current = this;
        let logText = '';
        let now = new Date();
        while (current = current._parentStateMachine) {
            logText = current._stateMachineName + '[' + current.currentState + '] | ' + logText;
        }
        let ticker = (now - this._tickerStamp);
        let unit = 'ms';
        if (ticker > 1000) {
            ticker = ticker / 1000;
            unit = 's';
        }
        logText += this._stateMachineName + ':[' + currentState + '(' + ticker + unit + ')]->[' + nextState + ']';
        if (args.length == 1) {
            args = args[0];
        }
        if (!args || (args instanceof Array && args.length == 0)) {
            Monitor.log(logText);
        }
        else {
            Monitor.log(logText);
            console.log('with:', args);
        }
    }

    elapse(states) {
        if (states) {
            let self = this;
            let time = 0;
            states.split('+').forEach(function (state) {
                self.stateStack.forEach(function (s) {
                    if (s.state == state) {
                        time += s.ticker;
                    }
                });
                if (state == self.currentState) {
                    time += (+new Date() - self._tickerStamp) / 1000;
                }
            });
            return time;

        }
        else {
            return (+new Date() - this._tickerStamp) / 1000;
        }

    }

    nextStateWhen(promise, state) {
        let self = this;
        let args = Array.prototype.slice.call(arguments, 1);
        promise.then(function () {
            self.nextState.apply(self, args);
        })
    }

    nextState(state) {
        if (this.currentState == state) {
            return;
        }
        let now = new Date().getTime();
        let args = Array.prototype.slice.call(arguments, 1);
        this.stateStack.push({state: this.currentState, ticker: (now - this._tickerStamp) / 1000});
        let currentState = this.currentState;
        this.currentState = state;
        this.log(currentState, state, args);
        this._ticker = 0;
        this._tickerStamp = now;
        for (let i = 0; i < this._breakPoint.length; i++) {//不封装是为了减少堆栈信息 便于断点调试
            let bp = this._breakPoint[i];
            if ((bp.currentState == '*' || bp.currentState === currentState) && ( bp == '*' || bp.nextState === state)) {
                debugger;
                break;
            }
        }
        let onMethodName = 'on' + _upperFirstCharCase(state),
            afterMethodName = 'after' + _upperFirstCharCase(state);
        this[onMethodName] && this[onMethodName].apply(this, args);
        this._emit.call(this, state, args);
        this._emit.call(this, '__sync__.' + state, args);
        this[afterMethodName] && this[afterMethodName].apply(this, args);
        if (this._destroyed) {
            this._destroy();
        }

    }

    syncState(target, targetState, thisState = targetState) {
        let self = this;
        target.on('__sync__.' + targetState, function (...args) {
            self.nextState(thisState, ...args);
        });
    }

    off(method, handler) {
        if (handler) {
            for (let i = this._eventHandlers[method].length; i-- > 0;) {
                if (this._eventHandlers[method][i] === handler) {
                    this._eventHandlers[method].splice(i, 1);
                    break;
                }
            }
        }
        else {
            delete this._eventHandlers[method];
        }
    }

    _emit(method, args) {
        let notEmpty = false;
        if (this._eventHandlers[method] && this._eventHandlers[method].length > 0) {
            notEmpty = true;
            for (let i = 0; i < this._eventHandlers[method].length; i++) {
                this._eventHandlers[method][i].apply(this, args);
            }
        }
        return this._parentStateMachine && this._parentStateMachine._emit(this._stateMachineName + '.' + method, args) || notEmpty;
    }

    emit(method) {
        let args = Array.prototype.slice.call(arguments, 1);
        return this._emit(method, args);
    }

    emitError(error, value) {
        if (!this.emit(error, value)) {
            this.emit('error', value, error);
        }
    }

    once(method, handler) {
        let self = this;
        let fired = false;

        function g() {
            this.off(method, g);
            if (!fired) {
                fired = true;
                handler.apply(self, Array.prototype.slice.call(arguments));
            }
        }

        this.on(method, g);
    }


    destroy() {
        this._destroyed = true;
    }

    _destroy() {
        this.currentState = 'destroy';
        for (let i = _stateMachines.length; i-- > 0;) {
            if (_stateMachines[i] === this) {
                _stateMachines.splice(i, 1);
                break;
            }
        }
        this._eventHandlers = {};
    }

    timeout(states, second, callback) {
        let self = this;
        states.split(',').forEach(function (state) {
            self._timeoutMonitor.push({
                match: new RegExp('^(' + state.split('+').join('|') + ')$'),
                timeout: second * 1000,
                callback: callback,
                elapse: 0
            });
        })

    }

    _checkTimeout(state, interval) {
        for (let i = 0; i < this._timeoutMonitor.length; i++) {
            let monitor = this._timeoutMonitor[i];
            if (monitor.match.test(state)) {
                monitor.elapse += interval;
            }
            if (monitor.elapse >= monitor.timeout) {
                monitor.callback && monitor.callback.call(this);
                monitor.elapse = 0;
            }
        }
        this._parentStateMachine && this._parentStateMachine._checkTimeout(this._stateMachineName + '.' + state, interval);
    }

    isInState(states) {
        let reg = new RegExp('^(' + states.replace(/,/g, '|') + ')$');
        return reg.test(this.currentState);
    }

    breakOn(stateSwitch) {
        let states = stateSwitch.split('->');
        this._breakPoint.push({currentState: states[0], nextState: states[1]});
    }
}
function _upperFirstCharCase(state) {
    return state.charAt(0).toUpperCase() + state.slice(1);
}
